/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.filecache;

import org.wso2.registry.RegistryException;
import org.wso2.registry.jdbc.filecache.FileData;
import org.wso2.registry.jdbc.filecache.RegistryFileInputStream;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.*;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Timer;

public class FileManager {

    private static Log log = LogFactory.getLog(FileManager.class);

    private static final long MAXIMUM_IDLE_TIME_FOR_FILE = 60 * 60 * 1000; // 1 hour

    private static final long PREFERRED_FILE_COUNT = 100;

    private static Map <String, FileData> fileDataMap = new HashMap <String, FileData> ();

    private static String tempDir;

    public static void init() {

        long oneHour = 60 * 60 * 1000;
        Timer timer = new Timer(true);
        timer.scheduleAtFixedRate(new FileCleanerTask(), oneHour, oneHour);
    }

    public static boolean hasFileBasedContent(String contentID) {

        if (fileDataMap.containsKey(contentID)) {
            return true;
        }

        return false;
    }

    public static InputStream getFileBasedInputStream(String contentID) throws RegistryException {

        String tempDir = getTempDirectory();
        String tempFilePath = tempDir + "reg" + contentID;
        File tempFile = new File(tempFilePath);
        if (!tempFile.exists()) {
            String msg = "Temporary content file is not created for content " + contentID;
            log.error(msg);
            throw new RegistryException(msg);
        }

        try {
            InputStream inputStream = new RegistryFileInputStream(
                    new BufferedInputStream(new FileInputStream(tempFile)), contentID);

            FileData fileData = fileDataMap.get(contentID);
            fileData.incrementConnection();

            return inputStream;

        } catch (FileNotFoundException e) {

            String msg = "Failed to create an input stream from the temp file " +
                    tempFilePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public static InputStream createFileBasedInputStream(String contentID, InputStream inputStream)
            throws RegistryException {

        if (inputStream == null) {
            return null;
        }

        String tempDir = getTempDirectory();
        String tempFilePath = tempDir + "reg" + contentID;
        File tempFile = new File(tempFilePath);
        if (!tempFile.exists()) {

            BufferedOutputStream out = null;

            try {
                tempFile.createNewFile();
                tempFile.deleteOnExit();

                // we increment the file data only after obtaining the input stream from file 
                FileData fileData = new FileData();
                fileDataMap.put(contentID, fileData);

                out = new BufferedOutputStream(new FileOutputStream(tempFile));
                byte[] contentChunk = new byte[1024];
                int byteCount;
                while ((byteCount = inputStream.read(contentChunk)) != -1) {
                    out.write(contentChunk, 0, byteCount);
                }
                out.flush();

            } catch (IOException e) {

                String msg = "Failed to write data to the temporary file " +
                        tempFilePath + ". " + e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg, e);

            } finally {

                try {
                    inputStream.close();

                    if (out != null) {
                        out.close();
                    }

                } catch (IOException e) {

                    String msg = "Failed to close streams used for temporary file. "
                            + e.getMessage();
                    log.error(msg, e);
                }

            }
        }


        try {
            InputStream fileBasedInputStream = new RegistryFileInputStream(
                    new BufferedInputStream(new FileInputStream(tempFile)), contentID);

            FileData fileData = fileDataMap.get(contentID);
            fileData.incrementConnection();

            return fileBasedInputStream;

        } catch (FileNotFoundException e) {

            String msg = "Failed to create an input stream from the temp file " +
                    tempFilePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public static File getFile(String contentID) {

        String tempDir = getTempDirectory();
        String tempFilePath = tempDir + "reg" + contentID;
        File tempFile = new File(tempFilePath);
        if (tempFile.exists()) {
            return tempFile;
        }

        return null;
    }

    public static synchronized void onStreamClosed(String contentID) {

        FileData fileData = fileDataMap.get(contentID);
        if (fileData != null) {
            fileData.decrementConnection();
            if (fileData.getConnections() <= 0 && fileDataMap.size() > PREFERRED_FILE_COUNT) {
                deleteFileBasedContent(contentID);
            }
        }
    }

    public static void cleanupFiles() {

        long currentTime = System.currentTimeMillis();

        Iterator <String> contentIDs = fileDataMap.keySet().iterator();

        while (contentIDs.hasNext()) {

            String contentID = contentIDs.next();

            FileData fileData = fileDataMap.get(contentID);
            if (fileData.getConnections() <= 0 &&
                    (currentTime - fileData.getIdleFrom()) >= MAXIMUM_IDLE_TIME_FOR_FILE) {

                deleteFileBasedContent(contentID);
            }
        }
    }

    public static void deleteFileBasedContent(String contentID) {

        fileDataMap.remove(contentID);

        File file = getFile(contentID);

        if (file != null) {
            file.delete();
        }
    }

    private static String getTempDirectory() {

        if (tempDir == null) {
            tempDir = System.getProperty("java.io.tmpdir");
            if (!tempDir.endsWith(File.separator)) {
                tempDir = tempDir + File.separator;
            }
        }

        return tempDir;
    }
}
